/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

/*

Header:

  WORD Identifier      Magic number. Always FFFFh
  WORD LengthOfHeader  Length of the header in 16-bit WORD
  WORD Version         Format version number
  WORD Transform       Image origin
  WORD Coords[4]       Size and position of image
  WORD PageSize[2]     Physical page size
  WORD Bounds[4]       Limits of coordinate system
  WORD Flags           Bit image opcode flag

*/

#define get16(a) le16toh(*(uint16_t*)(a))

static int
get16c(const void *a)
{
    int val = get16(a);
    if(val & 0x8000)
        val -= 0x10000;
    return val;
}

typedef struct{
    int x;
    int y;
}vertex_t;

typedef struct{
    void(*v_alpha_text)();
    void(*v_arc)();
    void(*v_bar)();
    void(*v_bit_image)();
    void(*v_circle)();
    void(*v_clear_disp_list)();
    void(*v_clrwk)();
    void(*v_ellarc)();
    void(*v_ellipse)();
    void(*v_ellpie)();
    void(*v_entercur)();
    void(*v_exitcur)();
    void(*v_fillarea)(void *userdata, int count, vertex_t *vertex);
    void(*v_form_adv)();
    void(*v_justified)();
    void(*v_line)();
    void(*v_output_window)();
    void(*v_pieslice)();
    void(*v_pmarker)();
    void(*v_qtext)();
    void(*v_rfbox)(void *userdata, vertex_t *vertex);
    void(*v_updwk)();
    void(*vr_recfl)();
    void(*vs_color)();
    void(*vsf_clip)();
    void(*vsf_color)(void *userdata, int color_index);
    void(*vsf_interior)();
    void(*vsf_perimeter)();
    void(*vsf_style)();
    void(*vsf_updat)();
    void(*vsl_color)(void *userdata, int color_index);
    void(*vsl_ends)(void *userdata, int beg_style, int end_style);
    void(*vsl_type)(void *userdata, int style);
    void(*vsl_udsty)();
    void(*vsl_width)();
    void(*vsm_color)(void *userdata, int color_index);
    void(*vsm_height)(void *userdata, vertex_t *vertex);
    void(*vsm_type)(void *userdata, int symbol);
    void(*vst_alignment)();
    void(*vst_color)();
    void(*vst_effects)();
    void(*vst_font)();
    void(*vst_height)();
    void(*vst_point)();
    void(*vst_rotation)(void *userdata, vertex_t *angle);
    void(*vswr_mode)();
}gem_vdi_interface_t;


static int
_gem_vdi_interpret(const uint8_t *p, size_t len, const gem_vdi_interface_t *vt, void *userdata)
{
    vertex_t vertex[4096];
    const uint8_t *end = p+len;
    int i;
    while(p+8 < end){
        int opcode = get16(p);
        int subcode = get16(p+6);
        int vertex_count = get16(p+2);
        int param_count = get16(p+4);
        switch(opcode){
        case 5: /* Alpha character cells */
            break;
        case 6: /* Polyline */
            /* v_line */
            break;
        case 9: /* Fill Area */
            //DEBUG_p("v_fillarea");
            for(i=0;i<vertex_count;++i){
                vertex[i].x = get16c(p+8 + i*4 + 0);
                vertex[i].y = get16c(p+8 + i*4 + 2);
                //DEBUG_p("%dx%d",vertex[i].x,vertex[i].y);
            }
            vt->v_fillarea(userdata, vertex_count, vertex);
            break;
        case 11:
            switch(subcode){
            case 1: /* Bar */
                /* v_bar */
                break;
            case 5: /* Ellipse */
                /* v_ellipse */
                break;
            case 6: /* Elliptical Arc */
                /* v_ellarc */
                break;
            case 7: /* Elliptical Pie Slice */
                /* v_ellpie */
                break;
            case 8: /* Rounded Rectangle */
                DEBUG_p("v_rbox");
                for(i=0;i<vertex_count;++i){
                    vertex[i].x = get16c(p+8 + i*4 + 0);
                    vertex[i].y = get16c(p+8 + i*4 + 2);
                    DEBUG_p("%dx%d",vertex[i].x,vertex[i].y);
                }
                /* vt->v_rbox(userdata, vertex); */
                break;
            case 9: /* Filled Rounded Rectangle */
                DEBUG_p("v_rfbox");
                for(i=0;i<vertex_count;++i){
                    vertex[i].x = get16c(p+8 + i*4 + 0);
                    vertex[i].y = get16c(p+8 + i*4 + 2);
                    DEBUG_p("%dx%d",vertex[i].x,vertex[i].y);
                }
                vt->v_rfbox(userdata, vertex);
                break;
            default:
                DEBUG_p("opcode: %d/%xh, subcode: %d/%xh, vertex count: %d, param count: %d",opcode, opcode, subcode, subcode, vertex_count,param_count);
            }
        //case 0x0b: /* Arc */
        //    break;
        case 13: /* Set Character Baseline Vector */
            for(i=0;i<vertex_count;++i){
                vertex[i].x = get16c(p+8 + i*4 + 0);
                vertex[i].y = get16c(p+8 + i*4 + 2);
                DEBUG_p("vsm_rotation: %dx%d",vertex[i].x,vertex[i].y);
            }
            /* vst_rotation(userdata, vertex); */
            break;
        case 15: /* Set Polyline Type */
            vt->vsl_type(userdata, get16(p + vertex_count*4));
            break;
        case 16: /* Set Polyline Width */
            /* FIXME!! */
            break;
        case 17: /* Set Polyline Color Index */
            vt->vsl_color(userdata, get16(p + vertex_count*4));
            break;
        case 18: /* Set Polymarker Type */
            /* vsm_type(userdata, get16(p + vertex_count*4); */
            break;
        case 19: /* Set Polymarker Height */
            for(i=0;i<vertex_count;++i){
                vertex[i].x = get16c(p+8 + i*4 + 0);
                vertex[i].y = get16c(p+8 + i*4 + 2);
                DEBUG_p("vsm_height: %dx%d",vertex[i].x,vertex[i].y);
            }
            /* vsm_height(userdata, vertex); */
            break;
        case 20: /* Set Polymarker Color Index */
            /* vt->vsm_color(userdata, get16(p + vertex_count*4)); */
            break;
            break;
        case 21: /* Set Text Font */
            /* vst_font */
            break;
        case 22: /* Set Text Color Index */
            /* vst_color */
            break;
        case 23: /* Set Fill Interior Style */
            /* vsf_interior */
            break;
        case 24: /* Set Fill Style Index */
            /* vsf_style */
            break;
        case 25: /* Set Fill Color Index */
            vt->vsf_color(userdata, get16(p + vertex_count*4));
            break;
        case 32: /* Set Writing Mode */
            /* FIXME!! */
            break;
        case 39: /* Set Graphic Text Alignment */
            /* vst_alignment */
            break;
        case 104: /* Set Fill Perimeter Visibility */
            /* vsf_perimeter */
            break;
        case 106: /* Set Graphics Text Effects */
            /* vst_effects */
            break;
        case 107: /* Set Character Height (points) */
            /* vst_point */
            break;
        case 108: /* Set Polyline End Styles */
            vt->vsl_ends(userdata, get16(p + vertex_count*4), get16(p + vertex_count*4 + 2));
            break;
        default:
            DEBUG_p("opcode: %d/%xh, subcode: %d/%xh, vertex count: %d, param count: %d",opcode, opcode, subcode, subcode, vertex_count,param_count);
            break;
        }
        p += 8 + vertex_count*4 + param_count*2;
    }
    return 0;
}

/* --- */

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    int xoffs;
    int yoffs;
    uint8_t *data;
    int len;
};


static const char **
_gv_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-gem-vdi",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_gv_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->scalable_size = 1;
    h->info->threadsafe = 1;
    return h;
}


static void
_gv_free(abydos_plugin_handle_t *h)
{
    free(h);
}


static int
_gv_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    int header_size;
    if(len < 16)
        goto error_truncated;
    if(get16(data) != 0xffff)
        goto error_invalid;

    header_size = get16(data+2)*2;
    DEBUG_p("header size: %d",header_size);

    DEBUG_p("Coords: %d, %d, %d, %d",get16c(data+8),get16c(data+10),get16c(data+12),get16c(data+14));
#if 1
    h->xoffs = get16c(data+8);
    h->yoffs = get16c(data+10);
    h->info->width = get16c(data+12) - h->xoffs;
    h->info->height = get16c(data+14) - h->yoffs;
#else
    h->info->width = (signed)get16c(data+12);
    h->info->height = (signed)get16c(data+14);
#endif
    DEBUG_p("image size: %dx%d",h->info->width,h->info->height);

    data += header_size;
    len -= header_size;

    h->data = malloc(len);
    memcpy(h->data, data, len);
    h->len = len;

    return 0;
error_invalid:
    DEBUG_p("invalid data");
    h->info->error = "Invalid data";
    return -1;
error_truncated:
    DEBUG_p("truncated data");
    h->info->error = "Truncated data";
    return -1;
}


typedef struct{
    cairo_t *cr;
}context_t;


static void
_set_color_from_index(cairo_t *cr, int index)
{
    DEBUG_p("color: %d",index);
    index %= 16;
    switch(index){
    case 0:
        cairo_set_source_rgb(cr, 0, 0, 0);
        break;
    case 2:
        cairo_set_source_rgb(cr, 1, 0, 0);
        break;
    case 3:
        cairo_set_source_rgb(cr, 0, 1, 0);
        break;
    case 4:
        cairo_set_source_rgb(cr, 0, 0, 1);
        break;
    case 5:
        cairo_set_source_rgb(cr, 0, 1, 1);
        break;
    case 6:
        cairo_set_source_rgb(cr, 1, 1, 0);
        break;
    case 7:
        cairo_set_source_rgb(cr, 1, 0, 1);
        break;
    case 8:
        cairo_set_source_rgb(cr, 2.0/3, 2.0/3, 2.0/3);
        break;
    case 9:
        cairo_set_source_rgb(cr, 1.0/3, 1.0/3, 1.0/3);
        break;
    case 10:
        cairo_set_source_rgb(cr, 2.0/3, 0, 0);
        break;
    case 11:
        cairo_set_source_rgb(cr, 0, 2.0/3, 0);
        break;
    case 12:
        cairo_set_source_rgb(cr, 2.0/3, 0, 0);
        break;
    case 13:
        cairo_set_source_rgb(cr, 0, 2.0/3, 2.0/3);
        break;
    case 14:
        cairo_set_source_rgb(cr, 2.0/3, 2.0/3, 0);
        break;
    case 15:
        cairo_set_source_rgb(cr, 2.0/3, 0, 2.0/3);
        break;
    default:
        cairo_set_source_rgb(cr, 0, 0, 0);
        break;
    }
}


static void
_on_v_fillarea(void *userdata, int count, vertex_t *v)
{
    context_t *c = (context_t*)userdata;
    int i;
#if 1
    cairo_move_to(c->cr, v[0].x, v[0].y);
    for(i=1;i<count;++i)
        cairo_line_to(c->cr, v[i].x, v[i].y);
    //cairo_set_source_rgb(c->cr,0,1,0);
    cairo_fill(c->cr);
#endif
}


static void
_on_v_rfbox(void *userdata, vertex_t *v)
{
    context_t *c = (context_t*)userdata;
    cairo_rectangle(c->cr, v[0].x, v[0].y, v[1].x-v[0].x, v[1].y-v[0].y);
    //cairo_set_source_rgb(c->cr,1,0,0);
    cairo_fill(c->cr);
}


static void
_on_vsf_color(void *userdata, int color_index)
{
    context_t *c = (context_t*)userdata;
    _set_color_from_index(c->cr, color_index);
}


static void
_on_vsl_color(void *userdata, int color_index)
{
    context_t *c = (context_t*)userdata;
    _set_color_from_index(c->cr, color_index);
}


static void
_on_vsl_ends(void *userdata, int beg_style, int end_style)
{
}


static void
_on_vsl_type(void *userdata, int style)
{
}


static void
_gv_render(abydos_plugin_handle_t *h,
           cairo_t *cr,
           int page,
           cairo_bool_t *layer_visibility,
           int variant,
           double time,
           int frame)
{
    static const gem_vdi_interface_t vt = {
        .v_fillarea = _on_v_fillarea,
        .v_rfbox = _on_v_rfbox,
        .vsf_color = _on_vsf_color,
        .vsl_color = _on_vsl_color,
        .vsl_ends = _on_vsl_ends,
        .vsl_type = _on_vsl_type
    };
    context_t c;
    c.cr = cr;
    cairo_translate(cr, -h->xoffs, -h->yoffs);
    //for(i=0;i<h->shape_count;++i)
    _gem_vdi_interpret(h->data, h->len, &vt, &c);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _gv_get_mime_types,
        .new = _gv_new,
        .free = _gv_free,
        .create_from_data = _gv_create_from_data,
        .render = _gv_render
    };
    return &p;
}
