/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <libmng.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */

#include "pixels.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif


typedef uint32_t pixel_t;

typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};

typedef struct{
    abydos_plugin_handle_t *h;
    const char *data;
    size_t data_len;
    int current_frame;
    pixel_t *pixels;
}read_context_t;


static const char **
_mng_get_mime_types()
{
    static const char *mime_types[] = {
        "video/x-mng",
        NULL
    };
    return mime_types;
}


static mng_ptr
_on_alloc(mng_size_t len)
{
    return calloc(1, len);
}


static void
_on_free(mng_ptr ptr, mng_size_t len)
{
    free(ptr);
}


static mng_bool
_return_true()
{
    return MNG_TRUE;
}


/* feed data to the decoder */
static mng_bool
_on_readstream(mng_handle mng, mng_ptr buffer, mng_uint32 size, mng_uint32 * bytesread)
{
    read_context_t *c = (read_context_t*)mng_get_userdata(mng);
    if(size > c->data_len)
        size = c->data_len;
    //DEBUG_p("copying %d bytes (of %d)",(int)size, (int)c->data_len);
    memcpy(buffer, c->data, size);
    c->data += size;
    c->data_len -= size;
    *bytesread = size;
    return MNG_TRUE;
}


/* the header's been read. set up the display stuff */
static mng_bool
_on_processheader(mng_handle mng, mng_uint32 width, mng_uint32 height)
{
    read_context_t *c = (read_context_t*)mng_get_userdata(mng);

    DEBUG_p("dimensions: %dx%d",width,height);
    c->h->info->width = width;
    c->h->info->height = height;

    DEBUG_p("width: %d",mng_get_imagewidth(mng));
    DEBUG_p("frame count: %d",mng_get_framecount(mng));
    DEBUG_p("total frames: %d",mng_get_totalframes(mng));
    DEBUG_p("layer count: %d",mng_get_layercount(mng));
    DEBUG_p("playtime: %d",mng_get_playtime(mng));

    c->h->info->frame_count = MAX(1,mng_get_framecount(mng));

    /* tell the mng decoder about our bit-depth choice */
    /* FIXME: this works on intel. is it correct in general? */
    mng_set_canvasstyle(mng, MNG_CANVAS_BGRA8);

    return MNG_TRUE;
}


/* return a row pointer for the decoder to fill */
static mng_ptr
_on_getcanvasline(mng_handle mng, mng_uint32 line)
{
    read_context_t *c = (read_context_t*)mng_get_userdata(mng);
    //DEBUG_p("line %d",line);
    return &c->pixels[c->h->info->width * line];
}


/* timer */
static mng_uint32
_on_getticks(mng_handle mng)
{
    DEBUG_p("begin");
    return 0;
}


static mng_bool
_on_settimer(mng_handle mng, mng_uint32 msecs)
{
    read_context_t *c = (read_context_t*)mng_get_userdata(mng);
    DEBUG_p("msec: %d",msecs);
    c->h->frame[c->current_frame].duration = (double)msecs/1000;
    return MNG_TRUE;
}


static mng_bool
_on_error(mng_handle mng, mng_int32 code, mng_int8 severity,
            mng_chunkid chunktype, mng_uint32 chunkseq,
            mng_int32 extra1, mng_int32 extra2, mng_pchar text)
{
    //read_context_t *c = (read_context_t*)mng_get_userdata(mng);
    union{
        char chunk[5];
        uint32_t num;
    }id;

    id.num = htobe64(chunktype);
    id.chunk[4] = '\0';

    fprintf(stderr, "error playing: chunk %s (%d):\n", id.chunk, chunkseq);
    fprintf(stderr, "%s\n", text);

    return MNG_FALSE;
}


static abydos_plugin_handle_t *
_mng_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->frame = NULL;
    return h;
}


static void
_mng_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static int
_mng_create_from_data(abydos_plugin_handle_t *h,const char *bytes,size_t len)
{
    mng_handle mng;
    read_context_t c;
    int i;

    mng = mng_initialize(&c, _on_alloc, _on_free, MNG_NULL);
    if(!mng)
        return -1;

    /* set the callbacks */
    mng_setcb_errorproc(mng, _on_error);
    mng_setcb_openstream(mng, _return_true);
    mng_setcb_closestream(mng, _return_true);
    mng_setcb_readdata(mng, _on_readstream);
    mng_setcb_gettickcount(mng, _on_getticks);
    mng_setcb_settimer(mng, _on_settimer);
    mng_setcb_processheader(mng, _on_processheader);
    mng_setcb_getcanvasline(mng, _on_getcanvasline);
    mng_setcb_refresh(mng, _return_true);

    c.h = h;
    c.data = bytes;
    c.data_len = len;
    mng_read(mng);

    DEBUG_p("frame count: %d",mng_get_framecount(mng));
    DEBUG_p("total frames: %d",mng_get_totalframes(mng));
    h->frame = (frame_t*)malloc(h->info->frame_count*sizeof(frame_t));

    c.pixels = (pixel_t*)malloc(h->info->width * h->info->height * sizeof(pixel_t));
    for(i=0;i<h->info->frame_count;++i){
        c.current_frame = i;
        mng_display(mng);
        h->frame[i].surface = abydos_image_surface_create_from_pixels(ABYDOS_PIXEL_FORMAT_b8g8r8a8, 0,
                                                                      (const uint8_t*)c.pixels, h->info->width, h->info->height,
                                                                      -1, 4, NULL);
    }
    free(c.pixels);

    mng_cleanup(&mng);
    return MNG_TRUE;
}


static cairo_surface_t *
_mng_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_mng_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _mng_get_mime_types,
        .new = _mng_new,
        .free = _mng_free,
        .create_from_data = _mng_create_from_data,
        .get_image_surface = _mng_get_image_surface,
        .get_duration = _mng_get_duration
    };
    return &p;
}
