/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <vips/vips.h>

#include "pixels.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static void
_vips_done()
{
    vips_shutdown();
}


static const char **
_vips_get_mime_types()
{
    static const char *mime_types[] = {
        "application/fits",
        "image/jpeg",
        "image/png",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_vips_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_vips_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static cairo_surface_t *
_surface_from_image(VipsImage *img)
{
    int width;
    int height;
    int channels = vips_image_get_bands(img);
    const uint8_t *src;
    abydos_pixel_format_t format;
    switch(vips_image_get_format(img)){
    case VIPS_FORMAT_UCHAR:
        switch(vips_image_get_interpretation(img)){
        case VIPS_INTERPRETATION_sRGB:
            format = channels >= 4 ? ABYDOS_PIXEL_FORMAT_r8g8b8a8 : ABYDOS_PIXEL_FORMAT_r8g8b8;
            break;
        default:
            return NULL;
        }
        break;
    default:
        return NULL;
    }
    width = vips_image_get_width(img);
    height = vips_image_get_height(img);
    src = vips_image_get_data(img);
    return abydos_image_surface_create_from_pixels(format, 0, src, width, height, -1, channels, NULL);
}


static int
_vips_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    VipsImage *img = vips_image_new_from_buffer(data, len, "", NULL);
    int res = -1;
    if(!img)
        return -1;
    if((h->surface = _surface_from_image(img))){
        h->info->pixel_ratio = vips_image_get_xres(img)/vips_image_get_yres(img);
        res = 0;
    }
    g_object_unref(img);
    return res;
}


static cairo_surface_t *
_vips_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .done = _vips_done,
        .get_mime_types = _vips_get_mime_types,
        .new = _vips_new,
        .free = _vips_free,
        .create_from_data = _vips_create_from_data,
        .get_image_surface = _vips_get_image_surface
    };
    VIPS_INIT("abydos");
    return &p;
}
